"""
Item Loader

See documentation in docs/topics/loaders.rst
"""
from contextlib import suppress

from itemadapter import ItemAdapter
from parsel.utils import extract_regex, flatten

from itemloaders.common import wrap_loader_context
from itemloaders.processors import Identity
from itemloaders.utils import arg_to_iter


def unbound_method(method):
    """
    Allow to use single-argument functions as input or output processors
    (no need to define an unused first 'self' argument)
    """
    with suppress(AttributeError):
        if "." not in method.__qualname__:
            return method.__func__
    return method


class ItemLoader:
    """
    Return a new Item Loader for populating the given item. If no item is
    given, one is instantiated automatically using the class in
    :attr:`default_item_class`.

    When instantiated with a :param ``selector`` parameter the :class:`ItemLoader` class
    provides convenient mechanisms for extracting data from web pages
    using parsel_ selectors.

    :param item: The item instance to populate using subsequent calls to
        :meth:`~ItemLoader.add_xpath`, :meth:`~ItemLoader.add_css`,
        :meth:`~ItemLoader.add_jmes` or :meth:`~ItemLoader.add_value`.
    :type item: :class:`dict` object

    :param selector: The selector to extract data from, when using the
        :meth:`add_xpath` (resp. :meth:`add_css`, :meth:`add_jmes`) or :meth:`replace_xpath`
        (resp. :meth:`replace_css`, :meth:`replace_jmes`) method.
    :type selector: :class:`~parsel.selector.Selector` object

    The item, selector and the remaining keyword arguments are
    assigned to the Loader context (accessible through the :attr:`context` attribute).

    .. attribute:: item

        The item object being parsed by this Item Loader.
        This is mostly used as a property so when attempting to override this
        value, you may want to check out :attr:`default_item_class` first.

    .. attribute:: context

        The currently active :ref:`Context <loaders-context>` of this Item Loader.
        Refer to <loaders-context> for more information about the Loader Context.

    .. attribute:: default_item_class

        An Item class (or factory), used to instantiate items when not given in
        the ``__init__`` method.

        .. warning:: Currently, this factory/class needs to be
            callable/instantiated without any arguments.
            If you are using ``dataclasses``, please consider the following
            alternative::

                from dataclasses import dataclass, field
                from typing import Optional

                @dataclass
                class Product:
                    name: Optional[str] = field(default=None)
                    price: Optional[float] = field(default=None)

    .. attribute:: default_input_processor

        The default input processor to use for those fields which don't specify
        one.

    .. attribute:: default_output_processor

        The default output processor to use for those fields which don't specify
        one.

    .. attribute:: selector

        The :class:`~parsel.selector.Selector` object to extract data from.
        It's the selector given in the ``__init__`` method.
        This attribute is meant to be read-only.

    .. _parsel: https://parsel.readthedocs.io/en/latest/
    """

    default_item_class = dict
    default_input_processor = Identity()
    default_output_processor = Identity()

    def __init__(self, item=None, selector=None, parent=None, **context):
        self.selector = selector
        context.update(selector=selector)
        if item is None:
            item = self.default_item_class()
        self._local_item = item
        context["item"] = item
        self.context = context
        self.parent = parent
        self._local_values = {}
        # values from initial item
        for field_name, value in ItemAdapter(item).items():
            self._values.setdefault(field_name, [])
            self._values[field_name] += arg_to_iter(value)

    @property
    def _values(self):
        if self.parent is not None:
            return self.parent._values
        else:
            return self._local_values

    @property
    def item(self):
        if self.parent is not None:
            return self.parent.item
        else:
            return self._local_item

    def nested_xpath(self, xpath, **context):
        """
        Create a nested loader with an xpath selector.
        The supplied selector is applied relative to selector associated
        with this :class:`ItemLoader`. The nested loader shares the item
        with the parent :class:`ItemLoader` so calls to :meth:`add_xpath`,
        :meth:`add_value`, :meth:`replace_value`, etc. will behave as expected.
        """
        selector = self.selector.xpath(xpath)
        context.update(selector=selector)
        subloader = self.__class__(item=self.item, parent=self, **context)
        return subloader

    def nested_css(self, css, **context):
        """
        Create a nested loader with a css selector.
        The supplied selector is applied relative to selector associated
        with this :class:`ItemLoader`. The nested loader shares the item
        with the parent :class:`ItemLoader` so calls to :meth:`add_xpath`,
        :meth:`add_value`, :meth:`replace_value`, etc. will behave as expected.
        """
        selector = self.selector.css(css)
        context.update(selector=selector)
        subloader = self.__class__(item=self.item, parent=self, **context)
        return subloader

    def add_value(self, field_name, value, *processors, re=None, **kw):
        """
        Process and then add the given ``value`` for the given field.

        The value is first passed through :meth:`get_value` by giving the
        ``processors`` and ``kwargs``, and then passed through the
        :ref:`field input processor <processors>` and its result
        appended to the data collected for that field. If the field already
        contains collected data, the new data is added.

        The given ``field_name`` can be ``None``, in which case values for
        multiple fields may be added. And the processed value should be a dict
        with field_name mapped to values.

        Examples::

            loader.add_value('name', 'Color TV')
            loader.add_value('colours', ['white', 'blue'])
            loader.add_value('length', '100')
            loader.add_value('name', 'name: foo', TakeFirst(), re='name: (.+)')
            loader.add_value(None, {'name': 'foo', 'sex': 'male'})
        """
        value = self.get_value(value, *processors, re=re, **kw)
        if value is None:
            return
        if not field_name:
            for k, v in value.items():
                self._add_value(k, v)
        else:
            self._add_value(field_name, value)

    def replace_value(self, field_name, value, *processors, re=None, **kw):
        """
        Similar to :meth:`add_value` but replaces the collected data with the
        new value instead of adding it.
        """
        value = self.get_value(value, *processors, re=re, **kw)
        if value is None:
            return
        if not field_name:
            for k, v in value.items():
                self._replace_value(k, v)
        else:
            self._replace_value(field_name, value)

    def _add_value(self, field_name, value):
        value = arg_to_iter(value)
        processed_value = self._process_input_value(field_name, value)
        if processed_value:
            self._values.setdefault(field_name, [])
            self._values[field_name] += arg_to_iter(processed_value)

    def _replace_value(self, field_name, value):
        self._values.pop(field_name, None)
        self._add_value(field_name, value)

    def get_value(self, value, *processors, re=None, **kw):
        """
        Process the given ``value`` by the given ``processors`` and keyword
        arguments.

        Available keyword arguments:

        :param re: a regular expression to use for extracting data from the
            given value using :func:`~parsel.utils.extract_regex` method,
            applied before processors
        :type re: str or typing.Pattern

        Examples:

        >>> from itemloaders import ItemLoader
        >>> from itemloaders.processors import TakeFirst
        >>> loader = ItemLoader()
        >>> loader.get_value('name: foo', TakeFirst(), str.upper, re='name: (.+)')
        'FOO'
        """
        if re:
            value = arg_to_iter(value)
            value = flatten(extract_regex(re, x) for x in value)

        for proc in processors:
            if value is None:
                break
            _proc = proc
            proc = wrap_loader_context(proc, self.context)
            try:
                value = proc(value)
            except Exception as e:
                raise ValueError(
                    "Error with processor %s value=%r error='%s: %s'"
                    % (_proc.__class__.__name__, value, type(e).__name__, str(e))
                ) from e
        return value

    def load_item(self):
        """
        Populate the item with the data collected so far, and return it. The
        data collected is first passed through the :ref:`output processors
        <processors>` to get the final value to assign to each item field.
        """
        adapter = ItemAdapter(self.item)
        for field_name in tuple(self._values):
            value = self.get_output_value(field_name)
            if value is not None:
                adapter[field_name] = value

        return adapter.item

    def get_output_value(self, field_name):
        """
        Return the collected values parsed using the output processor, for the
        given field. This method doesn't populate or modify the item at all.
        """
        proc = self.get_output_processor(field_name)
        proc = wrap_loader_context(proc, self.context)
        value = self._values.get(field_name, [])
        try:
            return proc(value)
        except Exception as e:
            raise ValueError(
                "Error with output processor: field=%r value=%r error='%s: %s'"
                % (field_name, value, type(e).__name__, str(e))
            ) from e

    def get_collected_values(self, field_name):
        """Return the collected values for the given field."""
        return self._values.get(field_name, [])

    def get_input_processor(self, field_name):
        proc = getattr(self, "%s_in" % field_name, None)
        if not proc:
            proc = self._get_item_field_attr(
                field_name, "input_processor", self.default_input_processor
            )
        return unbound_method(proc)

    def get_output_processor(self, field_name):
        proc = getattr(self, "%s_out" % field_name, None)
        if not proc:
            proc = self._get_item_field_attr(
                field_name, "output_processor", self.default_output_processor
            )
        return unbound_method(proc)

    def _get_item_field_attr(self, field_name, key, default=None):
        field_meta = ItemAdapter(self.item).get_field_meta(field_name)
        return field_meta.get(key, default)

    def _process_input_value(self, field_name, value):
        proc = self.get_input_processor(field_name)
        _proc = proc
        proc = wrap_loader_context(proc, self.context)
        try:
            return proc(value)
        except Exception as e:
            raise ValueError(
                "Error with input processor %s: field=%r value=%r "
                "error='%s: %s'"
                % (
                    _proc.__class__.__name__,
                    field_name,
                    value,
                    type(e).__name__,
                    str(e),
                )
            ) from e

    def _check_selector_method(self):
        if self.selector is None:
            raise RuntimeError(
                "To use XPath or CSS selectors, %s "
                "must be instantiated with a selector" % self.__class__.__name__
            )

    def add_xpath(self, field_name, xpath, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.add_value` but receives an XPath instead of a
        value, which is used to extract a list of strings from the
        selector associated with this :class:`ItemLoader`.

        See :meth:`get_xpath` for ``kwargs``.

        :param xpath: the XPath to extract data from
        :type xpath: str

        Examples::

            # HTML snippet: <p class="product-name">Color TV</p>
            loader.add_xpath('name', '//p[@class="product-name"]')
            # HTML snippet: <p id="price">the price is $1200</p>
            loader.add_xpath('price', '//p[@id="price"]', re='the price is (.*)')

        """
        values = self._get_xpathvalues(xpath, **kw)
        self.add_value(field_name, values, *processors, re=re, **kw)

    def replace_xpath(self, field_name, xpath, *processors, re=None, **kw):
        """
        Similar to :meth:`add_xpath` but replaces collected data instead of adding it.
        """
        values = self._get_xpathvalues(xpath, **kw)
        self.replace_value(field_name, values, *processors, re=re, **kw)

    def get_xpath(self, xpath, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.get_value` but receives an XPath instead of a
        value, which is used to extract a list of unicode strings from the
        selector associated with this :class:`ItemLoader`.

        :param xpath: the XPath to extract data from
        :type xpath: str

        :param re: a regular expression to use for extracting data from the
            selected XPath region
        :type re: str or typing.Pattern

        Examples::

            # HTML snippet: <p class="product-name">Color TV</p>
            loader.get_xpath('//p[@class="product-name"]')
            # HTML snippet: <p id="price">the price is $1200</p>
            loader.get_xpath('//p[@id="price"]', TakeFirst(), re='the price is (.*)')

        """
        values = self._get_xpathvalues(xpath, **kw)
        return self.get_value(values, *processors, re=re, **kw)

    def _get_xpathvalues(self, xpaths, **kw):
        self._check_selector_method()
        xpaths = arg_to_iter(xpaths)
        return flatten(self.selector.xpath(xpath, **kw).getall() for xpath in xpaths)

    def add_css(self, field_name, css, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.add_value` but receives a CSS selector
        instead of a value, which is used to extract a list of unicode strings
        from the selector associated with this :class:`ItemLoader`.

        See :meth:`get_css` for ``kwargs``.

        :param css: the CSS selector to extract data from
        :type css: str

        Examples::

            # HTML snippet: <p class="product-name">Color TV</p>
            loader.add_css('name', 'p.product-name')
            # HTML snippet: <p id="price">the price is $1200</p>
            loader.add_css('price', 'p#price', re='the price is (.*)')
        """
        values = self._get_cssvalues(css)
        self.add_value(field_name, values, *processors, re=re, **kw)

    def replace_css(self, field_name, css, *processors, re=None, **kw):
        """
        Similar to :meth:`add_css` but replaces collected data instead of adding it.
        """
        values = self._get_cssvalues(css)
        self.replace_value(field_name, values, *processors, re=re, **kw)

    def get_css(self, css, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.get_value` but receives a CSS selector
        instead of a value, which is used to extract a list of unicode strings
        from the selector associated with this :class:`ItemLoader`.

        :param css: the CSS selector to extract data from
        :type css: str

        :param re: a regular expression to use for extracting data from the
            selected CSS region
        :type re: str or typing.Pattern

        Examples::

            # HTML snippet: <p class="product-name">Color TV</p>
            loader.get_css('p.product-name')
            # HTML snippet: <p id="price">the price is $1200</p>
            loader.get_css('p#price', TakeFirst(), re='the price is (.*)')
        """
        values = self._get_cssvalues(css)
        return self.get_value(values, *processors, re=re, **kw)

    def _get_cssvalues(self, csss):
        self._check_selector_method()
        csss = arg_to_iter(csss)
        return flatten(self.selector.css(css).getall() for css in csss)

    def add_jmes(self, field_name, jmes, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.add_value` but receives a JMESPath selector
        instead of a value, which is used to extract a list of unicode strings
        from the selector associated with this :class:`ItemLoader`.

        See :meth:`get_jmes` for ``kwargs``.

        :param jmes: the JMESPath selector to extract data from
        :type jmes: str

        Examples::

            # HTML snippet: {"name": "Color TV"}
            loader.add_jmes('name')
            # HTML snippet: {"price": the price is $1200"}
            loader.add_jmes('price', TakeFirst(), re='the price is (.*)')
        """
        values = self._get_jmesvalues(jmes)
        self.add_value(field_name, values, *processors, re=re, **kw)

    def replace_jmes(self, field_name, jmes, *processors, re=None, **kw):
        """
        Similar to :meth:`add_jmes` but replaces collected data instead of adding it.
        """
        values = self._get_jmesvalues(jmes)
        self.replace_value(field_name, values, *processors, re=re, **kw)

    def get_jmes(self, jmes, *processors, re=None, **kw):
        """
        Similar to :meth:`ItemLoader.get_value` but receives a JMESPath selector
        instead of a value, which is used to extract a list of unicode strings
        from the selector associated with this :class:`ItemLoader`.

        :param jmes: the JMESPath selector to extract data from
        :type jmes: str

        :param re: a regular expression to use for extracting data from the
            selected JMESPath
        :type re: str or typing.Pattern

        Examples::

            # HTML snippet: {"name": "Color TV"}
            loader.get_jmes('name')
            # HTML snippet: {"price": the price is $1200"}
            loader.get_jmes('price', TakeFirst(), re='the price is (.*)')
        """
        values = self._get_jmesvalues(jmes)
        return self.get_value(values, *processors, re=re, **kw)

    def _get_jmesvalues(self, jmess):
        self._check_selector_method()
        jmess = arg_to_iter(jmess)
        if not hasattr(self.selector, "jmespath"):
            raise AttributeError(
                "Please install parsel >= 1.8.1 to get jmespath support"
            )
        return flatten(self.selector.jmespath(jmes).getall() for jmes in jmess)



import re
import unittest
from unittest.mock import MagicMock

from parsel import Selector

from itemloaders import ItemLoader
from itemloaders.processors import MapCompose, TakeFirst


class CustomItemLoader(ItemLoader):
    name_in = MapCompose(lambda v: v.title())


class SelectortemLoaderTest(unittest.TestCase):
    selector = Selector(
        text="""
    <html>
    <body>
    <div id="id">marta</div>
    <p>paragraph</p>
    <a href="http://www.scrapy.org">homepage</a>
    <img src="/images/logo.png" width="244" height="65" alt="Scrapy">
    </body>
    </html>
    """
    )

    jmes_selector = Selector(
        text="""
    {
      "name": "marta",
      "description": "paragraph",
      "website": {
        "url": "http://www.scrapy.org",
        "name": "homepage"
      },
      "logo": "/images/logo.png"
    }
    """
    )

    def test_init_method(self):
        loader = CustomItemLoader()
        self.assertEqual(loader.selector, None)

    def test_init_method_errors(self):
        loader = CustomItemLoader()
        self.assertRaises(RuntimeError, loader.add_xpath, "url", "//a/@href")
        self.assertRaises(RuntimeError, loader.replace_xpath, "url", "//a/@href")
        self.assertRaises(RuntimeError, loader.get_xpath, "//a/@href")
        self.assertRaises(RuntimeError, loader.add_css, "name", "#name::text")
        self.assertRaises(RuntimeError, loader.replace_css, "name", "#name::text")
        self.assertRaises(RuntimeError, loader.get_css, "#name::text")

    def test_init_method_with_selector(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)

        loader.add_xpath("name", "//div/text()")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])

    def test_init_method_with_selector_css(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)

        loader.add_css("name", "div::text")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])

        loader.add_css("url", "a::attr(href)")
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])

        # combining/accumulating CSS selectors and XPath expressions
        loader.add_xpath("name", "//div/text()")
        self.assertEqual(loader.get_output_value("name"), ["Marta", "Marta"])

        loader.add_xpath("url", "//img/@src")
        self.assertEqual(
            loader.get_output_value("url"),
            ["http://www.scrapy.org", "/images/logo.png"],
        )

    def test_add_xpath_re(self):
        loader = CustomItemLoader(selector=self.selector)
        loader.add_xpath("name", "//div/text()", re="ma")
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

        loader = CustomItemLoader(selector=self.selector)
        loader.add_xpath("name", "//div/text()", re=re.compile("ma"))
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

    def test_add_xpath_variables(self):
        loader = CustomItemLoader(selector=self.selector)
        loader.add_xpath("name", "id($id)/text()", id="id")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader = CustomItemLoader(selector=self.selector)
        loader.add_xpath("name", "id($id)/text()", id="id2")
        self.assertEqual(loader.get_output_value("name"), [])

    def test_replace_xpath(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)
        loader.add_xpath("name", "//div/text()")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_xpath("name", "//p/text()")
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

        loader.replace_xpath("name", ["//p/text()", "//div/text()"])
        self.assertEqual(loader.get_output_value("name"), ["Paragraph", "Marta"])

    def test_get_xpath(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertEqual(loader.get_xpath("//p/text()"), ["paragraph"])
        self.assertEqual(loader.get_xpath("//p/text()", TakeFirst()), "paragraph")
        self.assertEqual(loader.get_xpath("//p/text()", TakeFirst(), re="pa"), "pa")

        self.assertEqual(
            loader.get_xpath(["//p/text()", "//div/text()"]), ["paragraph", "marta"]
        )

    def test_replace_xpath_multi_fields(self):
        loader = CustomItemLoader(selector=self.selector)
        loader.add_xpath(None, "//div/text()", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_xpath(None, "//p/text()", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

    def test_replace_xpath_re(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)
        loader.add_xpath("name", "//div/text()")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_xpath("name", "//div/text()", re="ma")
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

    def test_add_css_re(self):
        loader = CustomItemLoader(selector=self.selector)
        loader.add_css("name", "div::text", re="ma")
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

        loader.add_css("url", "a::attr(href)", re="http://(.+)")
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])

        loader = CustomItemLoader(selector=self.selector)
        loader.add_css("name", "div::text", re=re.compile("ma"))
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

        loader.add_css("url", "a::attr(href)", re=re.compile("http://(.+)"))
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])

    def test_replace_css(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)
        loader.add_css("name", "div::text")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_css("name", "p::text")
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

        loader.replace_css("name", ["p::text", "div::text"])
        self.assertEqual(loader.get_output_value("name"), ["Paragraph", "Marta"])

        loader.add_css("url", "a::attr(href)", re="http://(.+)")
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])
        loader.replace_css("url", "img::attr(src)")
        self.assertEqual(loader.get_output_value("url"), ["/images/logo.png"])

    def test_get_css(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertEqual(loader.get_css("p::text"), ["paragraph"])
        self.assertEqual(loader.get_css("p::text", TakeFirst()), "paragraph")
        self.assertEqual(loader.get_css("p::text", TakeFirst(), re="pa"), "pa")

        self.assertEqual(
            loader.get_css(["p::text", "div::text"]), ["paragraph", "marta"]
        )
        self.assertEqual(
            loader.get_css(["a::attr(href)", "img::attr(src)"]),
            ["http://www.scrapy.org", "/images/logo.png"],
        )

    def test_replace_css_multi_fields(self):
        loader = CustomItemLoader(selector=self.selector)
        loader.add_css(None, "div::text", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_css(None, "p::text", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

        loader.add_css(None, "a::attr(href)", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])
        loader.replace_css(None, "img::attr(src)", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(loader.get_output_value("url"), ["/images/logo.png"])

    def test_replace_css_re(self):
        loader = CustomItemLoader(selector=self.selector)
        self.assertTrue(loader.selector)
        loader.add_css("url", "a::attr(href)")
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])
        loader.replace_css("url", "a::attr(href)", re=r"http://www\.(.+)")
        self.assertEqual(loader.get_output_value("url"), ["scrapy.org"])

    def test_jmes_not_installed(self):
        selector = MagicMock(spec=Selector)
        del selector.jmespath
        loader = CustomItemLoader(selector=selector)
        with self.assertRaises(AttributeError) as err:
            loader.add_jmes("name", "name", re="ma")

        self.assertEqual(
            str(err.exception), "Please install parsel >= 1.8.1 to get jmespath support"
        )

    def test_add_jmes_re(self):
        loader = CustomItemLoader(selector=self.jmes_selector)
        loader.add_jmes("name", "name", re="ma")
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

        loader.add_jmes("url", "website.url", re="http://(.+)")
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])

        loader = CustomItemLoader(selector=self.jmes_selector)
        loader.add_jmes("name", "name", re=re.compile("ma"))
        self.assertEqual(loader.get_output_value("name"), ["Ma"])

        loader.add_jmes("url", "website.url", re=re.compile("http://(.+)"))
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])

    def test_get_jmes(self):
        loader = CustomItemLoader(selector=self.jmes_selector)
        self.assertEqual(loader.get_jmes("description"), ["paragraph"])
        self.assertEqual(loader.get_jmes("description", TakeFirst()), "paragraph")
        self.assertEqual(loader.get_jmes("description", TakeFirst(), re="pa"), "pa")

        self.assertEqual(
            loader.get_jmes(["description", "name"]), ["paragraph", "marta"]
        )
        self.assertEqual(
            loader.get_jmes(["website.url", "logo"]),
            ["http://www.scrapy.org", "/images/logo.png"],
        )

    def test_replace_jmes(self):
        loader = CustomItemLoader(selector=self.jmes_selector)
        self.assertTrue(loader.selector)
        loader.add_jmes("name", "name")
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_jmes("name", "description")
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

        loader.replace_jmes("name", ["description", "name"])
        self.assertEqual(loader.get_output_value("name"), ["Paragraph", "Marta"])

        loader.add_jmes("url", "website.url", re="http://(.+)")
        self.assertEqual(loader.get_output_value("url"), ["www.scrapy.org"])
        loader.replace_jmes("url", "logo")
        self.assertEqual(loader.get_output_value("url"), ["/images/logo.png"])

    def test_replace_jmes_multi_fields(self):
        loader = CustomItemLoader(selector=self.jmes_selector)
        loader.add_jmes(None, "name", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Marta"])
        loader.replace_jmes(None, "description", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(loader.get_output_value("name"), ["Paragraph"])

        loader.add_jmes(None, "website.url", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])
        loader.replace_jmes(None, "logo", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(loader.get_output_value("url"), ["/images/logo.png"])

    def test_replace_jmes_re(self):
        loader = CustomItemLoader(selector=self.jmes_selector)
        self.assertTrue(loader.selector)
        loader.add_jmes("url", "website.url")
        self.assertEqual(loader.get_output_value("url"), ["http://www.scrapy.org"])
        loader.replace_jmes("url", "website.url", re=r"http://www\.(.+)")
        self.assertEqual(loader.get_output_value("url"), ["scrapy.org"])



import unittest

from itemloaders.processors import SelectJmes


class SelectJmesTestCase(unittest.TestCase):
    test_list_equals = {
        "simple": ("foo.bar", {"foo": {"bar": "baz"}}, "baz"),
        "invalid": ("foo.bar.baz", {"foo": {"bar": "baz"}}, None),
        "top_level": ("foo", {"foo": {"bar": "baz"}}, {"bar": "baz"}),
        "double_vs_single_quote_string": ("foo.bar", {"foo": {"bar": "baz"}}, "baz"),
        "dict": (
            "foo.bar[*].name",
            {"foo": {"bar": [{"name": "one"}, {"name": "two"}]}},
            ["one", "two"],
        ),
        "list": ("[1]", [1, 2], 2),
    }

    def test_output(self):
        for key in self.test_list_equals:
            expr, test_list, expected = self.test_list_equals[key]
            test = SelectJmes(expr)(test_list)
            self.assertEqual(
                test, expected, msg=f"test {key!r} got {test} expected {expected}"
            )
